#include <stdint.h>

#include <avr/io.h>

#include <util/delay.h>

#include "i2c.h"
#include "ds1307.h"
#include "date_time.h"
#include "bcd.h"

// Sets the time using the RTC.
uint8_t date_time_set_rtc(struct date_time_t* date_time) {
	
	uint8_t day, month;
	uint16_t year;	
	
	date_time_get_date_from_day_number(date_time->date + CPM_EPOCH, &year, &month, &day);
	uint8_t ds1307[] = {
		date_time->second,      // Seconds
		date_time->minute,      // Minutes
		date_time->hour,        // Hours
		0x01,                   // Day
		bcd_encode(day),        // Date
		bcd_encode(month),      // Month
		bcd_encode(year % 100), // Year
		_BV(DS1307_SQWE),       // Control
	};
	return i2c_write(DS1307_SLAVE_ADDRESS, 0, ds1307, sizeof(ds1307)) == 0;
}

// Gets the time from the RTC.
uint8_t date_time_get_rtc(struct date_time_t* date_time) {

	uint8_t ds1307[7];
	if (i2c_read(DS1307_SLAVE_ADDRESS, 0, ds1307, sizeof(ds1307))) {
		return 0x00;
	} else {
		date_time->second = ds1307[0];
		date_time->minute = ds1307[1];
		date_time->hour = ds1307[2];
		
		date_time->date = date_time_get_day_number_from_epoch(
			date_time_expand_year(bcd_decode(ds1307[6])),
			bcd_decode(ds1307[5]),
			bcd_decode(ds1307[4])
		);
		return 0xFF;
	}
}

// Checks if the RTC time is valid; if not, the clock is reset when booting.
uint8_t date_time_rtc_is_valid(void) {
	uint8_t ds1307[7];
	if (i2c_read(DS1307_SLAVE_ADDRESS, 0, ds1307, sizeof(ds1307))) {
		return 0;
	} else {
		
		// Check that each field is valid BCD.
		for (uint8_t field = 0; field < sizeof(ds1307); ++field) {
			if (!bcd_is_valid(ds1307[field])) return 0;
		}
		
		// Check each field is in range.
		if (bcd_decode(ds1307[0]) > 59) return 0; // Seconds
		if (bcd_decode(ds1307[1]) > 59) return 0; // Minutes
		if (bcd_decode(ds1307[2]) > 23) return 0; // Hours
		
		
		// Extract the date components from the RTC.
		uint16_t year_a = date_time_expand_year(bcd_decode(ds1307[6]));
		uint8_t month_a = bcd_decode(ds1307[5]);
		uint8_t day_a = bcd_decode(ds1307[4]);
		
		// Calculate the day number.
		uint32_t day_number = date_time_get_day_number(year_a, month_a, day_a);
		
		// Convert the day number back into a date.
		uint16_t year_b;
		uint8_t month_b, day_b;		
		date_time_get_date_from_day_number(day_number, &year_b, &month_b, &day_b);
		
		return (year_a == year_b) && (month_a == month_b) && (day_a == day_b);
		
	}
}