#include <stdint.h>

#include <avr/io.h>
#include <util/delay.h>

#include "mcp23s08.h"
#include "i2c.h"
#include "ds1307.h"

#include "parallel.h"

// IODIR: 1: Input, 0: Ouptut.
// GPPU: 1: Pull-up, 0: No pull-up.
// GPIO: 1: Logic high, 0: Logic low.
// OLAT: Output latch

uint8_t parallel_is_bidirectional;
uint8_t parallel_select_printer;
uint8_t parallel_status_output;

void parallel_init(void) {
	parallel_status_output = 0xFF;
	mcp23s08_write(0b11, MCP23S08_GPPU, 0xFF);
	mcp23s08_write(0b11, MCP23S08_IODIR, 0xFF);
	parallel_select_printer = !0;
	parallel_write(PARALLEL_DATA, 0xFF);
	parallel_write(PARALLEL_CONTROL, 0x0C);
}

void parallel_write(enum parallel_register_t address, uint8_t value) {
	switch (address) {
		case PARALLEL_DATA:
			if (parallel_is_bidirectional) {
				// If the parallel port is bidirectional, 0 bits are driven low and 1 bits are weakly pulled high.
				mcp23s08_write(0b10, MCP23S08_GPPU, value);
				mcp23s08_write(0b10, MCP23S08_IODIR, value);
			}
			// Set the output data.
			mcp23s08_write(0b10, MCP23S08_GPIO, value);
			break;
		case PARALLEL_STATUS:
			// Read-only.
			break;
		case PARALLEL_CONTROL:
			// Bi-directional mode?
			if (value & _BV(5)) {
				// Enable bi-directional mode.
				parallel_is_bidirectional = !0;
				uint8_t current_output = mcp23s08_read(0b10, MCP23S08_OLAT);
				// Current 0 bits are driven low and 1 bits are weakly pulled high.
				mcp23s08_write(0b10, MCP23S08_GPPU, current_output);
				mcp23s08_write(0b10, MCP23S08_IODIR, current_output);
			} else {
				// Disable bi-directional mode.
				parallel_is_bidirectional = 0;
				// All data pins are driven high or low.
				mcp23s08_write(0b10, MCP23S08_IODIR, 0);
				mcp23s08_write(0b10, MCP23S08_GPPU, 0);
			}
			// Update the control pins.
			// Select printer (inverted).
			if (value & _BV(3)) {
				if (parallel_select_printer) {
					parallel_select_printer = 0;
					uint8_t ds1307[] = { 0 };
					i2c_write(DS1307_SLAVE_ADDRESS, DS1307_REGISTER_CONTROL, ds1307, sizeof(ds1307));
				}
			} else {
				if (!parallel_select_printer) {
					parallel_select_printer = !0;
					uint8_t ds1307[] = { _BV(DS1307_OUT) };
					i2c_write(DS1307_SLAVE_ADDRESS, DS1307_REGISTER_CONTROL, ds1307, sizeof(ds1307));
				}
			}
			// Initialise printer.
			if (value & _BV(2)) {
				parallel_status_output |= _BV(PARALLEL_INITIALISE);
			} else {
				parallel_status_output &= (uint8_t)~_BV(PARALLEL_INITIALISE);
			}
			// Auto linefeed (inverted).
			if (value & _BV(1)) {
				parallel_status_output &= (uint8_t)~_BV(PARALLEL_AUTO_LINEFEED);
			} else {
				parallel_status_output |= _BV(PARALLEL_AUTO_LINEFEED);
			}
			// Strobe (inverted).
			if (value & _BV(0)) {
				parallel_status_output &= (uint8_t)~_BV(PARALLEL_STROBE);
			} else {
				parallel_status_output |= _BV(PARALLEL_STROBE);
			}
			// Update the I/O expander.
			mcp23s08_write(0b11, MCP23S08_GPPU, parallel_status_output);
			mcp23s08_write(0b11, MCP23S08_IODIR, parallel_status_output);
			mcp23s08_write(0b11, MCP23S08_GPIO, parallel_status_output);
			break;
	}

}

uint8_t parallel_read(enum parallel_register_t address) {
	switch (address) {
		case PARALLEL_DATA:
			if (parallel_is_bidirectional) {
				return mcp23s08_read(0b10, MCP23S08_GPIO);
			} else {
				return mcp23s08_read(0b10, MCP23S08_OLAT);
			}
			break;
		case PARALLEL_STATUS:
			{
				uint8_t current_status = mcp23s08_read(0b11, MCP23S08_GPIO);
				uint8_t result = 0xFF;
				// Busy (inverted).
				if (current_status & _BV(PARALLEL_BUSY)) {
					result &= (uint8_t)~_BV(7);
				}
				// Ack.
				if (!(current_status & _BV(PARALLEL_ACK))) {
					result &= (uint8_t)~_BV(6);
				}
				// Paper out.
				if (!(current_status & _BV(PARALLEL_PAPER_OUT))) {
					result &= (uint8_t)~_BV(5);
				}
				// Select.
				if (!(current_status & _BV(PARALLEL_SELECT))) {
					result &= (uint8_t)~_BV(4);
				}
				// Error.
				if (!(current_status & _BV(PARALLEL_ERROR))) {
					result &= (uint8_t)~_BV(3);
				}
				return result;
			}
			break;
		case PARALLEL_CONTROL:
			{
				uint8_t current_control = mcp23s08_read(0b11, MCP23S08_GPIO);
				uint8_t result = 0xFF;
				if (!parallel_is_bidirectional) {
					result &= (uint8_t)~_BV(5);
				}
				// Select printer is handled by the DS1307.
				// Initialise printer.
				if (!(current_control & _BV(PARALLEL_INITIALISE))) {
					result &= (uint8_t)~_BV(2);
				}
				// Auto linefeed (inverted).
				if (current_control & _BV(PARALLEL_AUTO_LINEFEED)) {
					result &= (uint8_t)~_BV(1);
				}
				// Strobe (inverted).
				if (current_control & _BV(PARALLEL_STROBE)) {
					result &= (uint8_t)~_BV(0);
				}
				return result;
			}
			break;
	}
	return 0xFF;
}