#include <avr/io.h>
#include <avr/pgmspace.h>

#include "sd.h"
#include "spi.h"

// Temporary storage for a complete SD card block (512 bytes).
uint8_t sd_block[512];

// CRC routines.
void sd_crc7(uint8_t* crc, uint8_t value, uint8_t bits);
void sd_crc16(uint16_t* crc, uint8_t value);

// Sends a command to the SD card with an argument and returns the response.
uint8_t sd_command(enum sd_command_t command, uint32_t argument) {
	
	// Dud.
	spi_transmit(0xFF);
	
	uint8_t crc = 0x00;

	// Command
	command |= 0x40;
	spi_transmit(command);
	sd_crc7(&crc, command, 8);

	// Argument
	for (uint8_t arg_byte_index = 0; arg_byte_index < 4; ++arg_byte_index) {
		uint8_t arg_byte = (uint8_t)(argument >> 24);
		spi_transmit(arg_byte);
		sd_crc7(&crc, arg_byte, 8);
		argument <<= 8;
	}
	
	// CRC
	sd_crc7(&crc, 0, 7);
	crc <<= 1;
	crc |= 1;
	spi_transmit(crc);
	
	// Dummy to allow the card to process command
	spi_transmit(0xFF);

	// Response
	return spi_transmit(0xFF);
}


// Resets and initialises the SD card.
enum sd_error_t sd_reset(void) {

	enum sd_error_t error = SD_ERROR_OK;
	
	SD_SS_DDR |= _BV(SD_SS);

	spi_set_speed(400000UL);
	
	// At least 80 clocks with the SS de-asserted.
	SD_SS_PORT |= _BV(SD_SS);
	for (uint8_t dud = 0; dud < 10; ++dud) spi_transmit(0xFF);
	
	// At least 80 clocks with the SS asserted.
	SD_SS_PORT &= (uint8_t)~_BV(SD_SS);
	for (uint8_t dud = 0; dud < 10; ++dud) spi_transmit(0xFF);
	
	// Reset the card
	sd_command(SD_CMD_GO_IDLE_STATE, 0);
	
	// Poll the card until it says that everything is reset by returning 0:
	{
		// Poll the card 5000 times.
		uint16_t timeout = 5000;
		while (timeout--) {
			sd_command(SD_CMD_APP_CMD, 0); // Use the SD-card specific version to ignore MMC cards.
			if (sd_command(SD_CMD_SEND_OP_COND, 0) == 0) break;
		}
		// Have we timed out?
		if (timeout == 0) {
			error = SD_ERROR_TIMEOUT;
			goto ret;
		}
	}
	
	// Full speed ahead!
	spi_set_speed(10000000UL);
	
	// Enable CRCs for reliability.
	if (sd_command(SD_CMD_CRC_ON_OFF, 1) != 0) {
		error = SD_ERROR_UNSPECIFIED;
	}
	
ret:

	SD_SS_PORT |= _BV(SD_SS);
	return error;
	
}

// Reads a 512 byte block from the specified address.
enum sd_error_t sd_read_block(uint32_t address, uint8_t* buffer) {
	
	enum sd_error_t error = SD_ERROR_OK;
	
	SD_SS_PORT &= (uint8_t)~_BV(SD_SS);
	
	// Use SET_BLOCKLEN to ensure 512 byte blocks.
	sd_command(SD_CMD_SET_BLOCKLEN, 512);
	
	// Begin reading at the specified address.
	sd_command(SD_CMD_READ_SINGLE_BLOCK, address);
	
	// Wait for data token.
	{
		uint16_t timeout = 5000;
		while (timeout--) {
			if (spi_transmit(0xFF) == 0xFE) break;
		}
		if (timeout == 0) {
			error = SD_ERROR_TIMEOUT;
			goto ret;
		}
	}
	
	// Used to calculate the CRC.
	uint16_t calculated_crc = 0;
	
	// Read 512 bytes.
	for (uint16_t a = 0; a < 512; ++a) {
		uint8_t value = spi_transmit(0xFF);
		buffer[a] = value;
		sd_crc16(&calculated_crc, value);
	}
	
	// CRC.
	uint16_t received_crc = (spi_transmit(0xFF) << 8) | spi_transmit(0xFF);
	
	if (received_crc != calculated_crc) {
		error = SD_ERROR_CRC;
	}

ret:
	
	SD_SS_PORT |= _BV(SD_SS);
	return error;
	
}

// Writes a 512 byte block to the specified address.
enum sd_error_t sd_write_block(uint32_t address, uint8_t* buffer) {

	SD_SS_PORT &= (uint8_t)~_BV(SD_SS);
	
	// Use SET_BLOCKLEN to ensure 512 byte blocks.
	sd_command(SD_CMD_SET_BLOCKLEN, 512);
	
	// Begin writing at the specified address.
	sd_command(SD_CMD_WRITE_BLOCK, address);

	// Delay for 16 clocks.
	spi_transmit(0xFF);
	spi_transmit(0xFF);
	
	// Send for data token.
	spi_transmit(0xFE);
	
	// CRC for the data.
	uint16_t crc = 0;
	
	// Write 512 bytes.
	for (uint16_t a = 0; a < 512; ++a) {
		uint8_t value = buffer[a];
		spi_transmit(value);
		sd_crc16(&crc, value);
	}
	
	// CRC.
	spi_transmit(crc >> 8);
	spi_transmit(crc);
	
	// Response.
	spi_transmit(0xFF);
	
	// Wait for idle at end.
	while (spi_transmit(0xFF) != 0xFF);
	
	SD_SS_PORT |= _BV(SD_SS);
	
	return SD_ERROR_OK;

}

// Reads a 128 byte record from the specified address.
enum sd_error_t sd_read_record(uint32_t address, uint8_t* buffer) {
	
	enum sd_error_t result = sd_read_block(address & 0xFFFFFE00UL, sd_block);
	if (result != SD_ERROR_OK) return result;
	
	uint16_t source = address & 0x1FF;
	for (uint8_t dest = 0; dest < 128; ++dest, ++source) {
		buffer[dest] = sd_block[source];
	}
	
	return SD_ERROR_OK;
}

// Reads the CSD from the card.
enum sd_error_t sd_read_csd(struct sd_csd_t* csd) {
	uint8_t* p = (uint8_t*)csd;
	p += sizeof(struct sd_csd_t);
	
	SD_SS_PORT &= (uint8_t)~_BV(SD_SS);
	
	sd_command(SD_CMD_SEND_CSD, 0);
	
	while (spi_transmit(0xFF) != 0xFE);
	
	for (uint8_t s = 0; s < sizeof(struct sd_csd_t); ++s) {
		*--p = spi_transmit(0xFF);
	}
	
	// CRC7
	spi_transmit(0xFF);	
	
	SD_SS_PORT |= _BV(SD_SS);

	return SD_ERROR_OK;

}

void sd_crc7(uint8_t* crc, uint8_t value, uint8_t bits) {
	while (bits--) {
		*crc <<= 1;
		if (value & _BV(7)) *crc |= 1;
		value <<= 1;
		if (*crc & _BV(7)) *crc ^= (uint8_t)0x09;
	}
	*crc &= (uint8_t)~_BV(7);
}

uint16_t sd_crc16_table[256] PROGMEM = {
	0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7, 0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
	0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6, 0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
	0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485, 0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
	0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4, 0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
	0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823, 0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
	0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12, 0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
	0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41, 0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
	0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70, 0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
	0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F, 0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
	0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E, 0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
	0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D, 0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
	0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C, 0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
	0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB, 0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
	0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A, 0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
	0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9, 0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
	0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8, 0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0
};

void sd_crc16(uint16_t* crc, uint8_t value) {
	*crc = (*crc << 8) ^ pgm_read_word(sd_crc16_table + (uint8_t)((*crc >> 8) ^ value));
}