#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>
#include <util/atomic.h>

#include "z80.h"
#include "z80_memory.h"

#include "pins.h"

// Initialise the Z80 utility routines.
void z80_init(void) {
	z80_memory_init();
	
	// Drive control pins that we always have access to.
	PIN_HIGH(Z80_BUSREQ);
	PIN_HIGH(Z80_WAIT_RESET);
	PIN_HIGH(Z80_INT);
	PIN_LOW(Z80_RESET);
	
	// Reset the /WAIT flip-flop.
	z80_reset_wait();
}

// Reset the Z80.
void z80_reset(void) {
	PIN_LOW(Z80_RESET);
	_delay_ms(1); // At least 3 Z80 clock cycles.
	PIN_HIGH(Z80_RESET);
}

// Take control of the bus.
void z80_control_bus(void) {
	
	// Asset /BUSREQ
	Z80_BUSREQ_PORT &= (uint8_t)~_BV(Z80_BUSREQ);
	
	// Wait for /BUSACK to indicate we have control of the bus.
	while (Z80_BUSACK_PIN & _BV(Z80_BUSACK));
	
	// Drive /WR, /RD, /MREQ and /IORQ instead of the Z80.
	PIN_HIGH(Z80_WR);
	PIN_HIGH(Z80_RD);
	PIN_HIGH(Z80_MREQ);
	PIN_HIGH(Z80_IORQ);
	
	// Take control of the address bus.
	z80_memory_control_address();
}

// Return control of the bus to the Z80.
void z80_release_bus(void) {

	// Release /WR, /RD, /MREQ and /IORQ to allow the Z80 to drive them.
	PIN_FLOAT(Z80_WR);
	PIN_FLOAT(Z80_RD);
	PIN_FLOAT(Z80_MREQ);
	PIN_FLOAT(Z80_IORQ);
	
	// Release the address bus.
	z80_memory_release_address();

	// Deassert /BUSREQ
	Z80_BUSREQ_PORT |= _BV(Z80_BUSREQ);
}


// Clear the wait condition.
void z80_reset_wait(void) {
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		Z80_DATA_DDR = 0x00;
		Z80_WAIT_RESET_PIN |= _BV(Z80_WAIT_RESET);
		Z80_WAIT_RESET_PIN |= _BV(Z80_WAIT_RESET);
	}
}

// Clear the wait condition and take control of the bus.
void z80_reset_wait_control_bus(void) {
	// Ensure that BUSREQ is asserted before resetting the wait condition.
	Z80_BUSREQ_PORT &= (uint8_t)~_BV(Z80_BUSREQ);
	z80_reset_wait();
	z80_control_bus();
}